#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (C) 2012  Hsin-Yi Chen

# Author(s): Hsin-Yi Chen <ossug.hychen@gmail.com>

# Permission is hereby granted, free of charge, to any person obtaining a
# copy of this software and associated documentation files (the "Software"),
# to deal in the Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish, distribute, sublicense,
# and/or sell copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
# DEALINGS IN THE SOFTWARE.
from mazo.extensionlib import ExtensionType
from mazo.missionlib import CreateMissionJob, DoMissionJob, FormatDataJob, ModifyMissionJob
from mazo.cmdlib import Command
from mazo import util

import re
import logging
import subprocess
import tempfile
import dialog
import yaml
import os
import datetime
from launchpadlib.launchpad import Launchpad

# -----------------------------------------------------------------------
# Global Variables
# -----------------------------------------------------------------------
LP_VALIDATE_BUGTASK_STATUS=("New",
                    "Incomplete",
                    "Invalid",
                    "Won't Fix",
                    "Confirmed",
                    "Triaged",
                    "In Progress",
                    "Fix Committed",
                    "Fix Released" )

LP_VALIDATE_BUGTASK_IMPORTANCE=(
                    'Critical',
                    'High',
                    'Medium',
                    'Low')

# -----------------------------------------------------------------------
# Utilitize Functions
# -----------------------------------------------------------------------
def get_bugtask(project, bug):
    "Get the bug's bug-task for this project"
    bugtasks = filter(lambda bt: bt.target == project, bug.bug_tasks)
    try:
        return bugtasks[0]
    except IndexError:
        return None

def download_bug_attachments(bug, output_dir):
    # write title and desc
    with open(os.path.join(output_dir, bug.title.encode('utf-8')), 'w') as descfd:
            descfd.write(bug.description)
    for attachment in bug.attachments:
        filename = os.path.join(output_dir, attachment.title)
        try:
            data = attachment.data.open()
            with open(filename, 'w') as f:
                f.write(data.read())
            data.close()
        except Exception, e:
            logging.debug(e)
            logging.error(
                    "Unable to download file {0}".format(attachment.title))

def m_extract_bugs(result):
    #@FIXME: check entry type
    if type(result).__name__ != 'generator':
        return [bt.bug for bt in result]
    else:
        return result

# userEditSTring and EditableData codes is wrote by  James Ferguson <james.ferguson@canonical.com>, modified by hychen
def userEditString(suffix = '',
                   startingString = "",
                   filenamePrefix='enter-comment-if-reqd.'):
    "Have user edit a string.  Return edited version"
    fid, filename = tempfile.mkstemp(prefix=filenamePrefix)
    os.write(fid, startingString.encode('utf-8') + suffix)
    os.close(fid)
    while True:
        p = subprocess.Popen(['dialog', '--editbox', filename, '80', '74'],stderr=subprocess.PIPE)
        if p.wait():
            exit()
        result = p.stderr.read()
        if not subprocess.call(['dialog', '--yesno', 'Confirm? \n'+result, '80', '74']):
            break
    logging.info("User input successfully received")
    os.remove(filename)

    # strip off the suffix - be flexible about leading whitespace
    suffix = suffix.lstrip()
    if (len(result) - len(suffix)) == result.find(suffix):
        result = result[:-len(suffix)]

    return result.lstrip()

# -----------------------------------------------------------------------
# GUI Helper Functions
# -----------------------------------------------------------------------
__DLG = dialog.Dialog()

def ask_which_bugtasks(sel_msg, confirm_msg, origbugtasks):
    return ask_which_entries(sel_msg, confirm_msg, origbugtasks,
                          tag=lambda bugtask: str(bugtask.bug.id),
                          transform_cb=lambda bugtask: (str(bugtask.bug.id), bugtask.bug.title, 1),
                          result_format_cb=lambda bugtask: u"{0}-{1}".format(bugtask.bug.id, bugtask.bug.title))

def ask_which_bugs(sel_msg, confirm_msg, origbugs):
    return ask_which_entries(sel_msg, confirm_msg, origbugs,
                          tag=lambda bug: str(bug.id),
                          transform_cb=lambda bug: (str(bug.id), bug.title, 1),
                          result_format_cb=lambda bug: u"{0}-{1}".format(bug.id, bug.title))

def exit_if_user_cancelled(msg, lines):
    if not ask_confirmation(msg, lines):
        exit()

def ask_which_entries(selection_msg, confirmation_msg, pooldata,
                      tag=None, transform_cb=None, result_format_cb=None):
    selected_tags = ask_selection(selection_msg, pooldata, transform_cb)
    if not selected_tags:
        print u"Canceld because no selected bug"
        exit()

    if tag:
        selected_data = filter(lambda e: tag(e) in selected_tags, pooldata)
    else:
        selected_data = filter(lambda e: e in selected_tags, pooldata)

    # if selected data is empty means logic error
    assert selected_data

    # prepare confirmation message
    confirmation_msg += '\n'
    if result_format_cb:
        confirmation_msg += '\n'.join(map(result_format_cb, selected_data))
    else:
        confirmation_msg += '\n'.join(selected_data)
    confirmation_msg += '\n'
    exit_if_user_cancelled(confirmation_msg, len(selected_data))
    return selected_data

def ask_confirmation(msg, lines=40):
    return __DLG.yesno(msg, 40+lines*2, 74, defaultno=True) == 0

def ask_selection(msg, choices, transform_cb=None):
    if transform_cb:
        choices = map(transform_cb, choices)
    count_choices = len(choices)
    (code, tags) = __DLG.checklist(msg, 20+count_choices*2, 74,
                                    10+count_choices*2, choices)
    # exit if cancel
    if code != 0:
        exit()
    return tags

# -----------------------------------------------------------------------
# Exception Classes
# -----------------------------------------------------------------------
class EntryNotFound(Exception):
    pass

# -----------------------------------------------------------------------
# Utilitize Classes
# -----------------------------------------------------------------------
class EditableData():
    """Class to manage editable data of some text and some NVPs.  Set
    baseText and namedValues, call userEdit, then examine them again."""
    def __init__(self, baseText = "", namedValues = {}):
        self.baseText = baseText
        self.namedValues = namedValues
    def userEdit(self):
        # generate the string and edit it
        nvpString = "".join(["%s: %s\n" % nvp for nvp in self.namedValues.iteritems()])
        text = userEditString('', nvpString + self.baseText)
        # extract the values again
        for name in self.namedValues.keys():
            match = re.search("^%s: *(.*)$" % name, text, re.MULTILINE)
            if match:
                self.namedValues[name] = match.group(1)
                text = (text[:match.start()] + text[match.end():]).lstrip()
        self.baseText = text

class Conditions(object):

    @classmethod
    ## \brief load a yaml file as a conditions object
    #  \param cls Condtions
    #  \param filename  str file name
    #  \return Conditions
    def load(cls, filename):
        with open(filename) as fd:
            return cls(yaml.load(fd))

    ## \brief constructor
    #  \param rawdata Conditions content
    def __init__(self, rawdata):
        self._rawdata = rawdata
        # fill sahred info
        UTCNOW = datetime.datetime.utcnow()
        win = datetime.timedelta(7)
        date = UTCNOW - win
        self.shared_info = {'date_before_now': date}

    ## \brief get the condition difinition
    #  \param condname str condition name
    #  \return dict
    def get(self, condname):
        try:
            cond = self._rawdata['conditions'][condname]
        except KeyError:
            return None
        else:
            for k in cond:
                v = cond[k]
                if '$' in v:
                    try:
                        v = self.shared_info[v.replace('$','')]
                    except KeyError:
                        # skip because this string does not need to be
                        # replaced
                        continue
                    cond[k] = v
            return cond

# -----------------------------------------------------------------------
# Job Classes
# -----------------------------------------------------------------------

## \breif A DoMissionJob to connect Launchpad before fetching data
#
class DoLPQueryJob(DoMissionJob):

    lp = None

    def do_prejob(self):
        self.connect_lp()
        super(DoLPQueryJob, self).do_prejob()

    def connect_lp(self):
        if not self.lp:
            system = os.getenv('LPSYSTEM') or 'production'
            self.logger.info("Connecting Launchpad {} server".format(system))
            cachedir = os.path.expanduser("~/.launchpadlib/cache")
            self.lp = Launchpad.login_with('lp-cli', system, cachedir)
        return self.lp

    def get_lp_entry(self, entry_type, entry_id, strict=True):
        if entry_type != 'people':
            entry_type = entry_type+'s'
        try:
            return getattr(self.lp, entry_type)[entry_id]
        except KeyError:
            if strict:
                raise EntryNotFound('Can not find {0} in Launchpad {1}'.format(entry_id, entry_type))
            else:
                return None

# -----------------------------------------------------------------------
# Extension Classes
# -----------------------------------------------------------------------

## \breif A Extension to create a mission message that maintask is to get
#         a entry from Launchpad.
#
# STDIN : None
# STDOUT: Encoded MessionMessage String
class ExtensionGet(object):
    __metaclass__ = ExtensionType
    responsible_for = CreateMissionJob
    interface = Command

    def custom_arguments(self):
        self.add_argument('entry_type',
                                choices=('bug', ),
                                help='sepecify entry type that search bugs in')
        self.add_argument('entry_id',
                                nargs='+',
                                help='sepcify entry id that search bugs in')

    def maintask(self, args):
        for entry_id in args.entry_id:
            ret = self.get_lp_entry(args.entry_type, entry_id, False)
            if ret:
                yield ret

## \breif A Extension to create a mission message that maintask is to
#         search bugtasks in a entry from Launchpad
#
# STDIN : None
# STDOUT: Encoded MessionMessage String
class ExtensionSearchbugtasks(object):
    __metaclass__ = ExtensionType
    responsible_for = CreateMissionJob
    interface = Command

    def custom_arguments(self):
        self.add_argument('entry_type',
                                choices=('people','project'),
                                help='sepecify entry type that search bugs in')
        self.add_argument('entry_id',
                                help='sepcify entry id that search bugs in')
        self.add_argument('--tag', dest='tags', action='append',
                                help='sepecify a bug tag')
        self.add_argument('--tags-combinator', dest='tags_combinator',
                                help='Search for any or all of the tags specified.',
                                choices=('Any', 'All'), default='Any')
        self.add_argument('--status', dest='status',
                                choices=LP_VALIDATE_BUGTASK_STATUS,
                                action='append',
                                help='sepecify bugtask status')
        self.add_argument('--importance',
                                choices=LP_VALIDATE_BUGTASK_IMPORTANCE,
                                action='append',
                                dest='importance',
                                help='sepecify bugtask importance')
        self.add_argument('--assignee', dest='assignee',
                                help='sepecify bug assignee')
        self.add_argument('--search-text', dest='search_text',
                                help='search text')
        self.add_argument('--milestone', dest='milestone', help='milestone name')
        self.add_argument('--package', dest='package', help='source package name (only for project)', action='append')

    def do_prejob(self):
        if self.cmd_args.package and self.cmd_args.entry_type != 'project':
            self.logger.error("package option is only for project entry")

    def maintask(self, args):
        entry = self.get_lp_entry(args.entry_type, args.entry_id)
        kwargs = self.transform_args(args, ('entry_id', 'entry_type', 'package'))
        if kwargs.get('assignee'):
            kwargs['assignee'] = self.get_lp_entry('people', [kwargs['assignee']])
        if kwargs.get('milestone'):
            kwargs['milestone'] = entry.getMilestone(name=kwargs['milestone'])

        if args.package and args.entry_type == 'project':
            ret = []
            for pkgname in args.package:
                srcpkg = entry.getSourcePackage(name=pkgname)
                ret.extend([e for e in srcpkg.searchTasks(**kwargs)])
            return ret
        else:
            return entry.searchTasks(**kwargs)

## \breif A Extension to create a mission message that modified mantask
#           arguments.
#
# STDIN : Encoded MissionMessage String
# STDOUT: Encoded MessionMessage String
class ExtensionApplyconds(object):
    __metaclass__ = ExtensionType
    responsible_for = ModifyMissionJob
    interface = Command

    def custom_arguments(self):
        self.add_argument('condition', help='condition name')
        self.add_argument('--config', help='conditions config')

    def new_maintask_args(self):
        if self.cmd_args.config:
            conds = Conditions.load(self.cmd_args.config)
        else:
            self.logger.error( 'condition config not found!')
            exit()
        try:
            cond = conds.get(self.cmd_args.condition)
        except KeyError:
            self.logger.error('can not find condition {0}'.format(self.cmd_args.condition))
        else:
            return cond.items()

## \breif A Extension to edit bugtasks collected from Launchpad
#
# STDIN : Encoded MissionMessage String
# STDOUT: None
class ExtensionEditbugtasks(object):
    __metaclass__ = ExtensionType
    responsible_for = DoLPQueryJob
    interface = Command

    def custom_arguments(self):
        self.add_argument('--status',
                      choices=LP_VALIDATE_BUGTASK_STATUS,
                      help='change status')
        self.add_argument('--assignee',
                      help='assigne to people')
        self.add_argument('--milestone',
                      help='milestone name')
        self.add_argument('-y', '--yes',
                      action='store_true',
                      help='auto anwser yes to question')
        self.add_argument('-m', '--message',
                      help='add a comment')

    def do_postjob(self):
        # process args
        kwargs = self.transform_args(self.cmd_args, ['yes', 'message'])

        # ask user to select bugtasks
        if self.cmd_args.yes:
            selected_bugtasks = self.result
        else:
            substr_update_attributes = []
            for k,v in kwargs.items():
                if not v:
                    continue
                substr_update_attributes.append(u"{0}:{1}".format(k,v))

            attr_msg = u"New attribute vlaues of each bugtasks:\n"
            attr_msg += u'\n'.join(substr_update_attributes)

            if self.cmd_args.message:
                attr_msg += "\nComment:{0}".format(self.cmd_args.message)

            sel_msg = u"Please select which bugtasks need to change attrbiutes\n\n"
            sel_msg += attr_msg

            confirmation_msg = u"Are you sure to modify these bugtasks?\n\n" + attr_msg
            selected_bugtasks = ask_which_bugtasks(sel_msg, confirmation_msg, self.result)

        for bugtask in selected_bugtasks:
            if self.cmd_args.message:
                print u"Add new comment"
                bugtask.bug.newMessage(content=self.cmd_args.message)
            if kwargs:
                self.editbugtaskattrs(bugtask, kwargs)
            bugtask.lp_save()
        print "Done!"

    def editbugtaskattrs(self, bugtask, kwargs):
        print u"Editing bugtask {0}".format(bugtask.title)
        for attr, attrval in kwargs.items():
            # skip if new attribute value is empty
            if attrval is None:
                continue

            # convert bug_target to a project
            target = self.lp.load(bugtask.target.self_link)

            # get old value
            orig_attr_val = getattr(bugtask, attr)

            # get new value
            if attr == 'assignee':
                attrval = self.get_lp_entry('people', attrval)
            if attr == 'milestone':
                milestone_name = attrval
                attrval = target.getMilestone(name=attrval)
                # Skip because milestone is not exists
                if not attrval:
                    print u"Skip to update {0}: Milestone {0} is not exists in {1}".format(milestone_name, target.name)
                    continue

            # skip if old and new attribute is same
            # otherwise update it
            if orig_attr_val == attrval:
                print u"Skip to update {0}: move {1} to {2}".format(attr, orig_attr_val, attrval)
            else:
                print u"Update {0}: move {1} to {2}".format(attr, orig_attr_val, attrval)
                setattr(bugtask, attr, attrval)

## \breif A Abstract Extension Class to format collected data
class ExtensionFormatdata(object):
    __metaclass__ = ExtensionType
    responsible_for = DoLPQueryJob, FormatDataJob
    interface = Command

    def custom_arguments(self):
        self.add_argument('target')

## \breif A Extension to print bugs collected from Launchpad
#
# STDIN : Encoded MissionMessage String
# STDOUT: string
class ExtensionPrintbugs(ExtensionFormatdata):

    def target_list(self):
        for bug in self.result:
            print u"Bug {0}: {1}".format(bug.id, bug.title)
            for task in bug.bug_tasks:
                print u" {0}: {1}".format(task.bug_target_name, task.status)

## \breif A Extension to print bugtasks collected from Launchpad
#
# STDIN : Encoded MissionMessage String
# STDOUT: string
class ExtensionPrintbugtasks(ExtensionFormatdata):

    def custom_arguments(self):
        super(ExtensionPrintbugtasks, self).custom_arguments()
        self.add_argument('--not-contain',
                      nargs=2,
                      metavar=('attribute_name','attribute_value'),
                      help='skip if the entry attribute does not contant sepcified value')
        self.add_argument('--contain',
                      nargs=2,
                      metavar=('attribute_name','attribute_value'),
                      help='only show if the entry attribute does not contant sepcified value')

    def target_list(self):
        specified_attr = None
        specified_value = None
        for bugtask in self.result:
            if self.cmd_args.not_contain:
                (specified_attr, specified_value) = self.cmd_args.not_contain
                # skip attr has specified value
                if specified_value not in util.get_obj_attrval(bugtask, specified_attr):
                    self.print_bugtask(bugtask)
            elif self.cmd_args.contain:
                (specified_attr, specified_value) = self.cmd_args.contain
                if specified_value in util.get_obj_attrval(bugtask, specified_attr):
                    self.print_bugtask(bugtask)
            else:
                 self.print_bugtask(bugtask)

    def print_bugtask(self, bugtask):
        print u"{} {} ({})".format(bugtask.bug.id, bugtask.bug.title, bugtask.web_link)

## \breif A Extension to duplicate bugtasks collected from Launchpad
#
# STDIN : Encoded MissionMessage String
# STDOUT: None
class ExtensionDupbugtasks(object):
    __metaclass__ = ExtensionType
    responsible_for = DoLPQueryJob
    interface = Command

    def custom_arguments(self):
        self.add_argument('--project', dest='project', action='append',
                          required=True,
                          help='add target project, at least one')
        self.add_argument('--status', dest='status',
                                choices=LP_VALIDATE_BUGTASK_STATUS,
                                help='sepecify bugtask status')
        self.add_argument('--assignee', dest='assignee')
        self.add_argument('--milestone', dest='milestone')
        self.add_argument('--insert-title', dest='insert_title',
                           default='',
                           help='insert string before bug title')
        self.add_argument('--append-title', dest='append_title',
                           default='',
                           help='append string after bug title')
        self.add_argument('--insert-desc', dest='insert_desc',
                           default='',
                           help='insert string before bug description')
        self.add_argument('--append-desc', dest='append_desc',
                           default='',
                           help='append string after bug description')
        self.add_argument('--add-tag', dest='add_tag', action='append')
        self.add_argument('--remove-tag', dest='remove_tag', action='append')

    def do_postjob(self):
        # process args
        prj = [self.get_lp_entry('project', prj_name) \
                for prj_name in self.cmd_args.project]
        self.target_project = prj[0]

        if self.cmd_args.milestone:
            self.target_milestone = self.target_project.getMilestone(name=self.cmd_args.milestone)
        else:
            self.target_milestone = None
        if self.cmd_args.assignee:
            self.target_assignee = self.get_lp_entry('people', self.cmd_args.assignee)
        else:
            self.target_assignee = None

        origbugs = m_extract_bugs(self.result)
        # ask user to decide which bugs need to be duplicated
        sel_msg = u'Do you want to duplicate these bug to {0}'.format(self.target_project.name)
        confirm_msg = u"You will duplicate theses bugs to {0}\n".format(self.target_project.name)
        selected_bugs = ask_which_bugs(sel_msg, confirm_msg, origbugs)

        # create bugtasks
        created_bugtasks = []
        for bug in selected_bugs:
            self.logger.info(u"Copying {0} to {1}".format(
                                                bug.title,
                                                self.target_project.name
                                                ))
            newbug = self.dupbug(self.target_project, bug)
            # set task values if new task was on old bug
            orig_bugtask = get_bugtask(self.target_project, bug)
            self.logger.debug(u"found old bug task {0}".format(orig_bugtask))
            new_bugtask = get_bugtask(self.target_project, newbug)
            self.logger.debug(u"found new bug task {0}".format(new_bugtask))
            self.set_bugtask_attr(orig_bugtask, new_bugtask)
            created_bugtasks.append(new_bugtask)
        self.logger.info(u"Done!")
        # show created bugtask url
        self.logger.info("\n".join([bugtask.web_link for bugtask in created_bugtasks]))

    def dupbug(self, target, origbug):
        """duplicate a bug in target
        """
        newtags = origbug.tags
        # add tags we want
        if self.cmd_args.add_tag:
            newtags += self.cmd_args.add_tag
        # remove tags we don't want
        if self.cmd_args.remove_tag:
            newtags = filter(lambda e: e not in self.cmd_args.remove_tag, origbug.tags)

        newtitle = '{0}{1}{2}'.format(self.cmd_args.insert_title,
                              origbug.title,
                              self.cmd_args.append_title)

        newdesc = '{0}{1}{2}'.format(self.cmd_args.insert_desc,
                              origbug.description,
                              self.cmd_args.append_desc)

        # ask user to modify editable data
        editable_data = self._ask_predit("%s\n\nCopied from bug #%d" % (newdesc, origbug.id), newtags, newtitle)

        newbug = self.lp.bugs.createBug(
                        title = editable_data.namedValues['Title'],
                        description = editable_data.baseText,
                        tags = editable_data.namedValues['Tags'],
                        private = origbug.private,
                        security_related = origbug.security_related,
                        target = target)
        self.logger.debug(u"Created a new bug {0}".format(newbug.id))
        # copy team subscriptions.  Could be not what's required where a
        # task is now gone, however, on Sutton project, team sub
        # disappears
        for person in [sub.person for sub in origbug.subscriptions]:
            if person.is_team:
                newbug.subscribe(person=person)
        return newbug

    def set_bugtask_attr(self, orig_bugtask, new_bugtask):
        try:
            # importance
            new_bugtask.importance = orig_bugtask.importance

            # status
            if self.cmd_args.status:
                new_bugtask.status = self.cmd_args.status
            else:
                new_bugtask.status = orig_bugtask.status

            # asignee
            new_bugtask.assignee = self.target_assignee

            # milestone
            new_bugtask.milestone = self.target_milestone

            new_bugtask.lp_save()
        except Exception, e :
            print e
            msg = u"Unable to set a bugtask's importance and status"
            self.logger.warning(msg)

    def _ask_predit(self, desc, tags, title):
        editableData = EditableData(desc,
            {"Projects" : " ".join(self.cmd_args.project),
             "Tags" : " ".join(tags),
             "Title" : title})
        editableData.userEdit()
        return editableData

## \breif A Extension to download attachments of bugtasks collected from Launchpad
#
# STDIN : Encoded MissionMessage String
# STDOUT: None
class ExtensionGrabbugattachments(object):
    __metaclass__ = ExtensionType
    responsible_for = DoLPQueryJob
    interface = Command

    def custom_arguments(self):
        self.add_argument('--output-dir', default=os.getcwd())

    def do_postjob(self):
        bugs = m_extract_bugs(self.result)

        # create sub directory
        self.output_dir = self.cmd_args.output_dir
        for bug in bugs:
            subdir_name = os.path.join(self.output_dir, str(bug.id))
            if not os.path.isdir(subdir_name):
                os.mkdir(subdir_name)
            # grab attachments
            self.logger.info(
                "Downloading attachments of the bug {0}".format(
                                                    bug.id))
            download_bug_attachments(bug, subdir_name)
        self.logger.info("Done!")
