## \file
# \author 2012 Hsin-Yi Chen
# \brief A module for communicating data between Classes
import string
import copy
import types
import pickle
import marshal
import yaml
import json

# -----------------------------------------------------------------------
# Utilitize Functions
# -----------------------------------------------------------------------

## \brief reconstruct function from bytecode
#  \param func_bytecode str marshaled bytecode of func_code
#  \param func_default dict func_defaults
#  \return Function
def reconstruct_function(func_bytecode, func_defaults):
    func_code = marshal.loads(func_bytecode)
    return types.FunctionType(func_code, globals(), func_code.co_name,
                             func_defaults)

## \brief Recorded All neccesery data for archiving mission
class MissionMessage(object):

    # \brief constructor
    # \param maintask_callback Function
    # \param miantask_args dict
    def __init__(self, maintask_callback, maintask_args):
        ## Main Task Callback
        self.maintask_callback = maintask_callback
        ## Main Task Defaults
        self.maintask_defaults = maintask_callback.func_defaults
        ## Main Task Callback Bytecode
        self.maintask_bytecode = marshal.dumps(maintask_callback.func_code)
        ## Main Task Arguments
        self.maintask_args = maintask_args

    ## \brief load MissionMessage from serialized file object
    #  \param  cls MissionMessage
    #  \param  fd file object
    #  \return MissionMessage
    @classmethod
    def load(cls, fd):
        return cls.loads(fd.read())

    ## \brief load MissionMessage from serialized string
    #  \param  cls MissionMessage
    #  \param  string encoded MissoinMessage string
    #  \return MissionMessage
    @classmethod
    def loads(cls, string):
        received = pickle.loads(string)
        callback = reconstruct_function(received['maintask_bytecode'],
                                            received['maintask_defaults'])
        obj = cls(callback, received['maintask_args'])
        return obj

    ## \brief dump serialized MissionMessage
    #  \return list
    def dump(self):
        output = {
            'maintask_bytecode': self.maintask_bytecode,
            'maintask_defaults': self.maintask_defaults,
            'maintask_args': self.maintask_args,
        }
        return pickle.dumps(output)

## \brief Interface for Job needs to process MissionMessage
#
# methods need to be implemented:
# - _send
# - _recieve method
class MissionJob(object):

    ## \brief recieve MissionMessage
    def recieve(self):
        return MissionMessage.loads(self._recieve())

    ## \brief send MissionMessage
    def send(self, message):
        self._send(message.dump())

## \brief Interface for Create MissionMessage
#
# To create a MissionMessage
#
# methods need to be implemented:
#   - maintask
class CreateMissionJob(MissionJob):

    def do_job(self):
        self.result = MissionMessage(
                            self.maintask,
                            self.maintask_args)
        self.send(self.result)

    def maintask(self, args):
        raise NotImplemented

    def _send(self, message):
        raise NotImplemented

    def _recieve(self):
        raise NotImplemented

## \brief Interface for Modify MissionMessage
#
# To recieve a MissionMessage and modify attributes of MissionMessage
#
# methods need to be implemented:
#   - new_maintask_args
class ModifyMissionJob(MissionJob):

    def do_job(self):
        self.recieved_message = self.recieve()
        self.modified_message = copy.copy(self.recieved_message)
        for argname, argval in self.new_maintask_args():
            setattr(self.modified_message.maintask_args, argname, argval)
        self.result = self.modified_message
        self.send(self.result)

    ## \brief provide new maintask args for replacing the maintask args of
    #         MissionMessage
    #  \return list ex. [(argname, argval), ...]
    def new_maintask_args(self):
        raise NotImplemented

    def _send(self, message):
        raise NotImplemented

    def _recieve(self):
        raise NotImplemented

## \brief Interface for doing generic task
#
# To execute maintask
#
# methods need to be implemented:
#   - maintask
class DoJob(object):

    def do_job(self):
        self.result = self.maintask(self.maintask_args)
        #@FIXME: This check need to move to another class
        if hasattr(self, 'recieved_message') and self.recieved_message and not self.result:
            self.logger.info(u"Entries Not Found!")
            exit()

    def maintask(self, args):
        raise NotImplemented

## \brief Interface for doing maintask of MissionMessage
#
# To recieve MissionMessage and execute maintask function of MissionMessage
class DoMissionJob(MissionJob, DoJob):

    def do_prejob(self):
        self.recieved_message = self.recieve()
        self.maintask = lambda args : self.recieved_message.maintask_callback(self, args)
        self.maintask_args = self.recieved_message.maintask_args

    def _send(self, message):
        raise NotImplemented

    def _recieve(self):
        raise NotImplemented

## \brief Interface for format the result created by DoMissionJob
#
#  required: self.cmd_args.target
#
class FormatDataJob(MissionJob):

    def do_postjob(self):
        method_name = 'target_{}'.format(self.cmd_args.target)
        try:
            method = getattr(self, method_name)
        except AttributeError:
            self.logger.error('can not find target {}'.format(method_name))
        else:
            method()

    ## \brief regsited targets
    # \return list
    def targets(self):
        exclude = ('targets')
        ret = filter(lambda e: e.startswith('target') and e not in exclude, dir(self))
        return map(lambda e: e.replace('target_',''), ret)

    ## \brief convert object to yaml format
    def toyaml(self, output):
        return yaml.dump(output)

    ## \brief convert object to json format
    def tojson(self, output):
        return json.dumps(output, sort_keys=True, indent=4)

    ## \brief convert object to pickle format
    def topickle(self, output):
        return pickle.dumps(output)

    def _send(self, message):
        raise NotImplemented

    def _recieve(self):
        raise NotImplemented

## \brief interface for formating metadata
#
#  required: self.cmd_args.target
class FormatMetaDataJob(FormatDataJob):

    ## delimiter for string.Template
    custom_tpl_delimiter = '##'

    def do_postjob(self):
        ## \brief composed metdata title
        self.title = self.cmd_args.title
        self._metadatas = {}
        for dataname, datapath in self.cmd_args.metadata:
            self._metadatas[dataname] = datapath

        self._templates = {}
        if self.cmd_args.template:
            for tplname, tplpath in self.cmd_args.template:
                self._templates[tplname] = tplpath

        super(FormatMetaDataJob, self).do_postjob()

    ## \brief get a metadata object
    #  \param name str metadata name
    #  \return dict if name is not provided
    #          metclass object if name is provided
    def metadatas(self, name=None):
        def _load(path):
            with open(path, 'r') as f:
                return yaml.load(f)
        return self._getattr('_metadatas', _load, name)

    ## \brief get a template object
    #  \param name str template name
    #  \param tpl_config dict custom attribute for string.Template
    #  \return dict if name is not provided
    #          string.Template if name is provied
    def templates(self, name=None, tpl_config=None):
        if tpl_config:
            tplcls = type('CustomTemplate', (string.Template,), tpl_config)
        else:
            tplcls = string.Template
        def _load(path):
            with open(path, 'r') as tplf:
                return tplcls(tplf.read())
        return self._getattr('_templates', _load, name)

    ## \brief get a attirbute with speical loader
    #  \param key str attribute name
    #  \param cb callback function
    #  \param name str a element key of attribute got by param `key`
    #  \return dict if name is not provided
    #          element of dict if name is provied
    def _getattr(self, key, cb, name=None):
        attr = getattr(self, key)
        if name:
            try:
                return cb(attr[name])
            except KeyError:
                return None
        return {name:cb(path) for name, path in attr.items()}

    ## \brief fake format data method
    def target_dummy(self):
        print 'dummy target'
